(require 'cl)
(require 'easymenu)
(require 'cleite-vars)

(defvar cleite:feeds:buffer-name "*cleite feeds*")

(defvar cleite:feeds:mode-map
  (let ((map (make-sparse-keymap)))
	(set-keymap-parent map cleite:global-map)
	(define-key map "q"  'cleite:quit)
	(define-key map "\r" 'cleite:feeds:select)
	(define-key map "s"  'cleite:feeds:refresh)
	(define-key map "N"  'cleite:feeds:go-to-next-unread)
	(define-key map "P"  'cleite:feeds:go-to-prev-unread)
	(define-key map "n"  'cleite:feeds:go-to-next)
	(define-key map "p"  'cleite:feeds:go-to-prev)
	(define-key map "a"  'cleite:feeds:add-feed)
	(define-key map "d"  'cleite:feeds:delete-feed)
	(define-key map "r"  'cleite:feeds:set-rank)
	(define-key map "O"  'cleite:feeds:manipulate-opml)
	map)
  "Keymap for cleite:feeds:mode")


(defvar cleite:feeds:mode-menu-spec
  '("Feeds"
	["Select feed"                 cleite:feeds:select           ]
	["Refresh feed list"           cleite:feeds:refresh          ]
	"----"
	["Select next unread feed"     cleite:feeds:go-to-next-unread]
	["Select previous unread feed" cleite:feeds:go-to-prev-unread]
	["Select next feed"            cleite:feeds:go-to-next       ]
	["Select previous feed"        cleite:feeds:go-to-prev       ]
	"----"
	["Add feed..."                 cleite:feeds:add-feed         ]
	["Delete feed..."              cleite:feeds:delete-feed      ]
	"----"
	["Manipulate OPML..."          cleite:feeds:manipulate-opml  ]
	"----"
	["Quit Cleite"                 cleite:quit                   ]))


(defvar cleite:feeds:mode-hook nil
  "Hook for cleite:feeds:mode")

(defvar cleite:feeds:list nil)

(easy-menu-define cleite:feeds:mode-menu
  cleite:feeds:mode-map
  "Menu used in cleite:feeds:mode"
  cleite:feeds:mode-menu-spec)


(defun cleite:feeds:count-unread-items (feed)
  (cdr (assq 'unread-items feed)))


(defun cleite:feeds:show-status (status)
  (cleite:feeds:on-feeds-buffer

	(if (or (null status)
			(= 0 (length status)))
		(setq mode-line-buffer-identification "Feeds")
	  (setq mode-line-buffer-identification (format "Feeds:%s" status)))

	(force-mode-line-update)
    (sit-for 0)))


;;;###autoload
(defun cleite:feeds:mode ()
  "Major mode for Cleite feeds view
\\{cleite:feeds:mode-map}"
  (interactive)
  (kill-all-local-variables)
  (setq major-mode 'cleite:feeds:mode)
  (setq mode-name "Cleite Feeds")
  (cleite:feeds:show-status nil)
  (use-local-map cleite:feeds:mode-map)
  (easy-menu-add cleite:feeds:mode-menu)
  (run-hooks 'cleite:feeds:mode-hook))


;;;###autoload
(defun* cleite:feeds (&key refresh)
  "feeds バッファがあれば呼び出し、無ければ作る。

REFRESH が t の場合は、バッファ全體を常に再構築する。
フィード URI 文字列の場合は、該當する行を常に再構築する。"
  (if (get-buffer cleite:feeds:buffer-name)
	  (progn
		(switch-to-buffer cleite:feeds:buffer-name)
		(delete-other-windows)

		(cond ((eq t refresh)
			   (cleite:feeds:refresh))

			  ((stringp refresh)
			   (cleite:feeds:rebuild-line refresh))))
	
	(switch-to-buffer (get-buffer-create cleite:feeds:buffer-name))
	(cleite:feeds:mode)
	(delete-other-windows)
	(cleite:feeds:refresh)))


(defun* cleite:feeds:refresh (&key feed-uri)
  (interactive)
  
  (cond (feed-uri
		 ;; 更新前と後を比べて unread-items の有無が變化した場合は、こ
		 ;; のフィードだけと言はず全體を再ソートする。
		 (let* ((pos                (cleite:feeds:feed-position feed-uri))
				(feed               (nth pos cleite:feeds:list))
				(unread-flag-before (< 0 (cdr (assq 'unread-items feed))))
				
				(fresh-feed         (cleite:call 'feed:get-info
												 `((feed-uri . ,feed-uri))))
				(unread-flag-after  (< 0 (cdr (assq 'unread-items fresh-feed)))))
		   
		   (replace cleite:feeds:list
					(list fresh-feed)
					:start1 pos :end1 (1+ pos))

		   (if (eq unread-flag-before unread-flag-after)
			   (cleite:feeds:rebuild-line feed-uri)
			 (cleite:feeds:rebuild-buffer))))

		(t
		 (cleite:feeds:show-status "Loading feed list...")

         (let ((need-desc (if cleite:feeds:display-feed-description
                              'true 'false)))
           (setq cleite:feeds:list (cleite:call 'feed:list
                                                `((need-desc . ,need-desc)))))
		 (cleite:feeds:rebuild-buffer)

		 (cleite:feeds:show-status nil))))


(defun cleite:feeds:sort-comparator (a b)
  (block exit
	(labels ((read-p (feed)
			     (= 0 (cdr (assq 'unread-items feed))))
			 
			 (date-of (feed)
				 (cdr (assq 'lastmod feed)))

			 (title-of (feed)
				 (cdr (assq 'title feed)))

			 (rank-of (feed)
				 (cdr (assq 'rank feed)))

			 (negate-if-descending (bool order)
				 (if (or (eq order 'descending)
						 (eq order 'read-first))
					 (not bool)
				   bool)))

	  (dolist (criteria cleite:feeds:sort-criteria)
		(let ((type  (car criteria))
			  (order (cdr criteria)))

		  (cond ((eq type 'read-flag)
				 (if (not (eq (read-p a) (read-p b)))
					 (return-from exit (negate-if-descending
										   (and (not (read-p a)) (read-p b)) order))))

				((eq type 'date)
				 (if (not (= (date-of a) (date-of b)))
					 (return-from exit (negate-if-descending
										   (< (date-of a) (date-of b)) order))))

				((eq type 'title)
				 (if (not (equal (title-of a) (title-of b)))
					 (return-from exit (negate-if-descending
										   (> 0 (compare-strings (title-of a) 0 nil
																 (title-of b) 0 nil)) order))))

				((eq type 'rank)
				 (if (not (equal (rank-of a) (rank-of b)))
					 (return-from exit (negate-if-descending
										   (< (rank-of a) (rank-of b)) order))))

				(t
				 ;; 不正な type
				 (error "invalid criterion type: %s" type)))))

	  ;; 結局決まらなかったので兩者は等しい事にする。
	  nil)))


(defun cleite:feeds:sort ()
  (setq cleite:feeds:list
		(sort cleite:feeds:list #'cleite:feeds:sort-comparator)))


(defun cleite:feeds:rebuild-buffer ()
  (cleite:feeds:on-feeds-buffer
	(let ((saved-point (point)))
	  
	  (setq buffer-read-only nil)
	  (erase-buffer)

	  (cleite:feeds:sort)
  
	  (dolist (feed cleite:feeds:list)
	
		(unless (= (point) (point-min))
		  (insert "\n"))

		(cleite:feeds:insert-line feed))
  
	  (setq buffer-read-only t)
	  (goto-char saved-point))))


(defun cleite:feeds:rebuild-line (feed-uri)
  (cleite:feeds:on-feeds-window
	(setq buffer-read-only nil)
	
	(save-excursion
	  (let* ((line-num    (cleite:feeds:feed-position feed-uri))
			 (feed        (nth line-num cleite:feeds:list))
			 
			 (begin-point (progn
							(goto-char (point-min))
							(forward-line line-num)
							(point)))
			 
			 (end-point   (progn
							(end-of-line)
							(point))))
		
		;; この行端から行末までを再構築する。
		(delete-region begin-point end-point)
		(cleite:feeds:insert-line feed)))

	(setq buffer-read-only t)))


(defun cleite:feeds:insert-line (feed)
  (let* ((title        (cdr (assq 'title        feed)))
		 (feed-uri     (cdr (assq 'feed-uri     feed)))
		 (rank         (cdr (assq 'rank         feed)))
		 (status       (cdr (assq 'status       feed)))
		 (all-items    (cdr (assq 'all-items    feed)))
		 (unread-items (cdr (assq 'unread-items feed)))
		   
		 (default-face (if (= unread-items 0)
						   'cleite:face:read-feed
						 'cleite:face:unread-feed))
		   
		 (title-face   (if (equal feed-uri (cleite:items:current-feed-uri))
						   'cleite:face:selected-feed
						 default-face))
		   
		 (prev         (point)))

    ;; タイトルに改行コードが入ってゐると不都合なので、改行コードは削除
    ;; する。
    (setq title (replace-regexp-in-string "\n" "" title))

	;; [Unread/All] (Rank) {Status} Title
	  
	(if (= unread-items 0)
		(insert (cleite:faced-string (format "[%3d/%3d]"
											 unread-items all-items)
									 'cleite:face:read-feed))

	  (insert (cleite:faced-string "["
								   'cleite:face:read-feed))
	  (insert (cleite:faced-string (format "%3d" unread-items)
								   'cleite:face:unread-feed))
	  (insert (cleite:faced-string (format "/%3d]" all-items)
								   'cleite:face:read-feed)))

	(insert (cleite:faced-string (format " (%d)" rank) title-face))
	  
	(unless (eq status 'ok)
	  (insert (cleite:faced-string (format " {%s}" status)
								   'cleite:face:error)))
	  
	(insert (cleite:faced-string (format " %s" title) title-face))
	  
	;; 今バッファに插入した部分についてテキスト屬性を設定する。
	(put-text-property prev (point) 'mouse-face 'highlight)
	(put-text-property prev (point) 'cleite:feed feed)
    
    (when (assq 'description feed)
      (lexical-let* ((descr-src    (cdr (assq 'description feed)))
                     (description  (replace-regexp-in-string "\n" "" descr-src))
                     (buffer       (current-buffer))
                     (show-message (lambda (old-point new-point)
                                     ;; 自分がこのバッファ上に居るので
                                     ;; なければ表示しない。…事にした
                                     ;; かったのだが、どうもさうなって
                                     ;; ゐない。
                                     (when (and cleite:feeds:display-feed-description
                                                (eq buffer (current-buffer)))
                                       (message "%s" description)))))
        
        (put-text-property prev (point) 'help-echo     description)
        (put-text-property prev (point) 'point-entered show-message)))))


(defun cleite:feeds:kill-buffer ()
  
  (let ((buf (get-buffer cleite:feeds:buffer-name)))
	(when buf
	  (delete-windows-on buf)
	  (kill-buffer buf)))

  (setq cleite:feeds:list nil))


(defmacro cleite:feeds:with-pointed-feed (feed-sym &rest rest)
  (declare (indent 1))
  `(block nil
	 (beginning-of-line)
	 (let ((,feed-sym (get-text-property (point) 'cleite:feed)))

	   (if (null ,feed-sym)
		   ;; カーソルがフィード上でない場所にあるので何もせず終了
		   (return nil))

	   ,@rest)))


(defmacro cleite:feeds:on-feeds-window (&rest rest)
  (declare (indent 0))
  `(with-selected-window
	   (get-buffer-window cleite:feeds:buffer-name)
	 ,@rest))


(defmacro cleite:feeds:on-feeds-buffer (&rest rest)
  (declare (indent 0))
  `(with-current-buffer cleite:feeds:buffer-name
     ,@rest))


(defun cleite:feeds:feed-position (feed-uri &optional list)
  (position-if (lambda (feed)
				 (let ((uri (cdr (assq 'feed-uri feed))))
				   (equal uri feed-uri)))
			   (or list cleite:feeds:list)))


(defun cleite:feeds:select ()
  (interactive)
  (cleite:feeds:with-pointed-feed
	  feed

	;; 撰擇されてゐるフィードがあればバッファを更新
	(let ((feed-uri (cleite:items:current-feed-uri)))
	  (when feed-uri
		(cleite:items:current-feed-uri :clear t)
		(cleite:feeds:rebuild-line feed-uri)))
	
	;; items ウインドウを開く
	(let ((feed-uri (cdr (assq 'feed-uri feed))))
	  (cleite:items feed-uri)
	  (cleite:feeds:rebuild-line feed-uri))))


(defun cleite:feeds:go-to-next ()
  (interactive)
  (unless (cleite:feeds:find-next-feed)
	(message "This is the last feed.")))


(defun cleite:feeds:go-to-prev ()
  (interactive)
  (unless (cleite:feeds:find-next-feed :reverse t)
	(message "This is the first feed.")))


(defun cleite:feeds:go-to-next-unread ()
  (interactive)
  (unless (cleite:feeds:find-next-feed :unread-only t)
	(message "This is the last unread feed.")))


(defun cleite:feeds:go-to-prev-unread ()
  (interactive)
  (unless (cleite:feeds:find-next-feed :unread-only t :reverse t)
	(message "This is the first unread feed.")))


(defun* cleite:feeds:find-next-feed (&key unread-only reverse)
  (cleite:feeds:on-feeds-window
	(cleite:feeds:with-pointed-feed
		feed
	  (let* ((feed-uri (cdr (assq 'feed-uri feed)))
			 (list     (if reverse
						   (reverse cleite:feeds:list)
						 cleite:feeds:list))
			 (pos      (cleite:feeds:feed-position feed-uri list))
			 (rest     (nthcdr (1+ pos) list))
			 (direc    (if reverse
						   -1 1)))

        (block nil
          ;; 檢索対象がもう無ければ失敗。
          (dolist (feed rest nil)
            (if unread-only
                ;; 未讀項目が存在すれば成功。しなければ次へ。
                (when (< 0 (cleite:feeds:count-unread-items feed))
                  (progn
                    (forward-line direc)
                    (return t)))
              ;; 常に成功
              (progn
                (forward-line direc)
                (return t)))))))))


(defun cleite:feeds:add-feed ()
  (interactive)
  (let ((uri (read-string "Enter URL of a feed or an HTML: ")))

	;; feed:subscribe してみる。feed-error が發生した場合は
	;; feed:discover する。
	(condition-case e
		(cleite:feeds:subscribe uri)
	  
	  (srpc:failure
	   (let* ((data (cdr e)))

		 (if (and (consp data)
				  (eq 'feed-error (car data)))
			 (cleite:feeds:discover-feeds uri)

		   ;; feed-error でないので再度投げる
		   (signal 'srpc:failure data)))))))


(defun cleite:feeds:subscribe (feed-uri)
  (message "Subscribing to the feed...")
  (cleite:call 'feed:subscribe
			   `((feed-uri . ,feed-uri)))
  (cleite:feeds:refresh)
  (message "done."))


(defun cleite:feeds:discover-feeds (uri)
  (message "It was not an RSS feed. Trying to discover feed URLs in it...")
  
  (let ((feeds (cleite:call 'feed:discover
							`((uri . ,uri)))))
	(cond ((null feeds)
		   (message "It was not an RSS feed nor an HTML that points to one."))

		  ((= 1 (length feeds))
		   (let* ((feed     (car feeds))
				  (feed-uri (cdr (assq 'feed-uri feed)))
				  (title    (cdr (assq 'title    feed)))
				  (choice (cleite:read-char-with-retry
						   (if title
							   (format (concat "It was not an RSS feed but was an HTML that points to one. "
											   "Do you want to subscribe to the feed named `%s'? (y/n): ")
									   title)
							 (format (concat "It was not an RSS feed but was an HTML that points to one. "
											 "Do you want to subscribe to it? (y/n): ")))
						   '(?y ?n))))
			 
			 (if (eq choice ?y)
				 (cleite:feeds:subscribe feed-uri))))

		  (t
		   ;; 二つ以上のフィードが見付かったので、一時的なバッファを作って選択肢を表示。
		   (let ((choice (cleite:choose-element
						  (concat "It was not an RSS feed but was an HTML that points to some feeds. "
								  "Please select a feed to subscribe or cancel by C-g: ")
						  (mapcar (lambda (feed)
									(let ((feed-uri (cdr (assq 'feed-uri feed)))
										  (title    (cdr (assq 'title    feed))))
									  
									  (if title
										  (list (format "Title : %s" title)
												(format "URL   : %s" feed-uri))
										
										(list (format "URL : %s" feed-uri)))))
								  feeds))))
			 
			 (let* ((feed     (nth choice feeds))
					(feed-uri (cdr (assq 'feed-uri feed))))

			   (cleite:feeds:subscribe feed-uri)))))))


(defun cleite:feeds:delete-feed ()
  (interactive)
  (cleite:feeds:with-pointed-feed feed
	(let ((title    (cdr (assq 'title feed)))
		  (feed-uri (cdr (assq 'feed-uri   feed))))
	  
	(when (yes-or-no-p (format "Do you really want to delete the feed `%s'? "
							   title))
		
	  (cleite:call 'feed:unsubscribe
				   `((feed-uri . ,feed-uri)))
	  (cleite:feeds:refresh)))))


(defun cleite:feeds:set-rank ()
  (interactive)
  (cleite:feeds:with-pointed-feed feed
	(let ((feed-uri (cdr (assq 'feed-uri feed))))

	  (cleite:feeds:set-rank-of feed-uri))))


(defun cleite:feeds:set-rank-of (feed-uri)
  (let* ((pos    (cleite:feeds:feed-position feed-uri cleite:feeds:list))
		 (feed   (nth pos cleite:feeds:list))
		 (title  (cdr (assq 'title feed)))
		 (choice (cleite:read-char-with-retry
				  (format "Set the rank of `%s': [1-5]" title)
				  '(?1 ?2 ?3 ?4 ?5))))
	
	(cleite:call 'feed:set-rank
				 `((feed-uri . ,feed-uri)
				   (rank     . ,(- choice ?0))))
	(cleite:feeds:refresh)))


(defun cleite:feeds:import-feed-list (feeds)
  (block nil
	;; ダイアログバッファにフィード一覽を表示して、
	;; yes-or-no-p で確認を取る。
	(cleite:with-temp-window "*OPML Content*"
	  (insert "Content of the OPML:\n")
	  (insert "\n")
	  (dolist (feed feeds)
		(let ((feed-uri   (cdr (assq 'feed-uri feed)))
			  (title-cell      (assq 'title    feed)) )
				   
		  (if title-cell
			  (insert (format "* %s\n" (cdr title-cell)))

			(insert (format "* %s (no title)\n" (feed-uri))))))
				 
	  (setq buffer-read-only t)
				 
	  (unless (yes-or-no-p "Do you really want to import this feed list? ")
		(return)))
	
	;; フィードを一つずつ feed:subscribe しながら、失敗したものだけ集め
	;; ておく。後で失敗したフィード一覽をダイアログに表示する。
	(let ((n-feeds      (length feeds))
		  (failed-feeds '()           ))

	  (loop for i from 1 to n-feeds do
			(let* ((feed       (nth (1- i) feeds)         )
				   (feed-uri   (cdr (assq 'feed-uri feed)))
				   (title-cell      (assq 'title    feed) ))

			  (message "[%3d/%3d] Subscribing to `%s'..."
					   i n-feeds (if title-cell
									 (cdr title-cell)
								   feed-uri))

			  (condition-case e
				  (cleite:call 'feed:subscribe
							   `((feed-uri . ,feed-uri)))
	  
				(srpc:failure
				 (setq failed-feeds `(,@failed-feeds ,feed))))))

	  (cleite:feeds:refresh)

	  (if (null failed-feeds)
		  (message "Imported the whole list successfully.")

		(message "%d/%d feeds could not be subscribed."
				 (length failed-feeds)
				 (length feeds))
		
		(cleite:with-dialog-buffer
		  
		  (insert "This is the list of feeds that were not subscribed due to an error. ")
		  (insert "You may retry importing the same OPML later if you want to do so.\n")
		  (insert "\n")

		  (insert "Type `q' to close this window.\n")
		  (insert "\n")
		  
		  (fill-region (point-min) (point-max))

		  (insert "\n")
		  
		  (dolist (feed failed-feeds)
			(let ((feed-uri   (cdr (assq 'feed-uri feed)))
				  (title-cell      (assq 'title    feed) ))
			  
			  (if title-cell
				  (progn
					(insert (format "* %s\n" (cdr title-cell)))
					(insert (format "  %s\n" feed-uri)))

				(insert (format "* %s\n" feed-uri)))

			  (insert "\n"))))))))
	

(defun cleite:feeds:manipulate-opml ()
  (interactive)
  (let ((choice (cleite:read-char-with-retry
				 "[e]xport OPML to file / Import OPML from [f]ile / Import OPML from [u]RL: "
				 '(?e ?f ?u))))

	(cond ((eq choice ?e)
		   (let* ((fpath (read-file-name "Enter a file name to export OPML: "
                                         nil   ;; dir
                                         nil   ;; default-filename
                                         nil)) ;; mustmatch
                  (opml  (cleite:call 'opml:export)))
             (with-temp-file fpath
               (insert opml))
             (message "Wrote %s" fpath)))

		  ((eq choice ?f)
		   (let* ((fpath (read-file-name "Select an OPML file to import: "
										 nil ;; dir
										 nil ;; default-filename
										 t)) ;; mustmatch
				  (opml  (with-temp-buffer
						   (insert-file-contents fpath)
						   (buffer-string)))
				  (feeds (cleite:call 'opml:parse
									  `((opml . ,opml)))))

			 (cleite:feeds:import-feed-list feeds)))
			 
		  ((eq choice ?u)
		   (let* ((opml-uri (read-string "Enter URL of an OPML: "))
				  (feeds    (cleite:call 'opml:parse
										 `((uri . ,opml-uri)))))

			 (cleite:feeds:import-feed-list feeds))))))


(provide 'cleite-feeds)